"""
Utilities methods that may be used for database querying purpose.

Because UUIDv7 are timestamped ordered and monotonically increasing,
there are a good solution for generating primary keys.

The design of UUIDv7 feet well with the design of BTree.

Because they contains a date time, a UUID range can be compute
in order to retrieve UUIDs generated at a given time.

"""

from datetime import UTC, date, datetime, time, timedelta
from uuid import UUID

__all__ = [
    "uuid7_bounds_from_datetime",
    "uuid7_bounds_from_date",
]


def _datetime_to_uuid7_lowest(dt: datetime) -> UUID:
    unix_ts_ms = int(dt.timestamp() * 1000)
    version = 0x07
    var = 2
    final_bytes = unix_ts_ms.to_bytes(6)
    final_bytes += (version << 12).to_bytes(2)
    final_bytes += ((var << 62) + 0x3000000000000000).to_bytes(8)
    return UUID(bytes=final_bytes)


def uuid7_bounds_from_datetime(
    dt_lower: datetime,
    dt_upper: datetime | None = None,
) -> tuple[UUID, UUID]:
    """
    Get uuid bound for a half-open interval.

    This function can be usefull to search for any rows based on a uuid7 in a sql query.
    If one parameter is set, then the search is based on a millisecond, because uuid7
    are only millisecond precision.

    The returned bound are half open, so the upper bound, from the ``dt_upper`` will
    not include in the result, only the first value to be excluded.

    If the the second parameter is ommited, then the bound only contains a millisecond,
    of dt_lower.

    :param dt_lower: the included left bound of the range.
    :param dt_upper: the excluded right bound of the range.
    """
    return _datetime_to_uuid7_lowest(dt_lower), _datetime_to_uuid7_lowest(
        dt_upper or (dt_lower + timedelta(milliseconds=1))
    )


def uuid7_bounds_from_date(dt: date, tz=UTC) -> tuple[UUID, UUID]:
    """
    Get uuid bound for a particular day.

    This function can be usefull to search for any rows based on a uuid7 in a sql query.
    The right bound return is the first uuid of the next day that should be excluded.

    :param dt: the included left bound of the range.
    :param tz: the timezone used to compute the UUID, it should always be ommited.
    """
    return _datetime_to_uuid7_lowest(
        datetime.combine(dt, time=time(tzinfo=tz))
    ), _datetime_to_uuid7_lowest(
        datetime.combine(dt + timedelta(days=1), time=time(tzinfo=tz))
    )



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import tomllib

sys.path.insert(0, os.path.abspath("../../src"))


# -- Project information -----------------------------------------------------
def _get_project_meta():
    with open("../../pyproject.toml", "rb") as pyproject:
        return tomllib.load(pyproject)["project"]


pkg_meta = _get_project_meta()
project = str(pkg_meta["name"])
author = f"{pkg_meta['authors'][0]['name']} <{pkg_meta['authors'][0]['email']}>"
copyright = f"{pkg_meta['authors'][0]['name']} and contributors"

# The short X.Y version
version = str(pkg_meta["version"])
# The full version, including alpha/beta/rc tags
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "autodoc2",
]

autodoc2_packages = [
    {"path": "../../lastuuid"},
]

autodoc2_module_all_regexes = [
    "lastuuid",
    r"lastuuid\.factories",
    r"lastuuid\.dummies",
    r"lastuuid\.utils",
]

autodoc2_docstring_parser_regexes = [
    # this will render all docstrings as Markdown
    (r".*", "myst"),
]

autodoc2_hidden_objects = ["undoc", "private", "inherited"]
autodoc2_render_plugin = "myst"

autodoc2_output_dir = "develop"
autodoc2_sort_names = True

# Configure MyST parser
myst_enable_extensions = [
    "colon_fence",
    "html_admonition",
    "html_image",
    "replacements",
    "smartquotes",
    "substitution",
    "tasklist",
    "deflist",
    "fieldlist",
    "linkify",
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []  # "_static"]
